#include <fstream>
#include "DataType.h"
#include "ErrorCode.h"
#include "ESA.h"
#include "StringKernel.h"
#include <sys/time.h>
#include <cmath>

using namespace std;

void help()
{
        printf("==================================================================\n"
               "=    Compute string kernel matrix as libsvm precomputed kernel   =\n"
               "==================================================================\n"
               "Usage: ./file-libsvmkm  swf  param  dataset_fn  kernelmatrix_fn\n"
               "Options:\n"
               "swf (substring weighting functions)\n"
               "  0 - Constant Weight (a dummy parameter)\n"
               "  1 - Exponential Decay (lambda=0.5)\n"
               "  2 - K-Spectrum (k=5)\n"
               "  3 - Bounded Range (n=1)\n"
               "param (swf parameter)\n\n"
               "Note: \n"
               "1. Every line of the dataset is considered as one example (<label><white_space><filename>).\n"
               "   The filenames correspond to some text files and all lines in each text file are\n"
               "   concatenated into one string with '\\n' replaced by ' '\n"
               "2. A SENTINEL character ('\\n') is appended to each example string\n\n"          
              );               
}


int main(int argc, char *argv[])
{
        if(argc != 5) 
        { 
                help(); 
                exit(EXIT_FAILURE);
        }

        int swf = atoi(argv[1]);
        Real swfParam = atof(argv[2]);
        string datasetfn = string(argv[3]);        
        string kernelmatrixfn = string(argv[4]);
        UInt32 datasize = 0;
        vector<string> data_vec;
        vector<UInt32> data_len;
        vector<int>   data_label;
        double *km = 0;
          
        ifstream indatafile;
        indatafile.open(datasetfn.c_str());
        if(!indatafile)
        {
                printf("Unable to open dataset <%s>\n", datasetfn.c_str());
                exit(EXIT_FAILURE);
        }

        printf("Reading data");
        int total = 0;
        string line;
        string filename;
        ifstream eachfile;
        string tmpline;
        int label = 0;
        int i = 0;       
        
        while(!indatafile.eof())
        {                
                filename.clear();
                indatafile >> label;
                getline(indatafile, filename);
		if(filename.size() < 1) continue;
		string::size_type const start = filename.find_first_not_of("\t ");
		string::size_type const  end = filename.find_last_not_of("\t ");
                filename = filename.substr(start,end-start+1);;  // trim leading/trailing whitespaces
		eachfile.open(filename.c_str());
                if(!eachfile.good()) 
                {
                        printf("\nUnable to open example file <%s>\n", filename.c_str());
                        exit(EXIT_FAILURE);
                }
    
                line.clear();                
                while(!eachfile.eof()) 
                {
                        tmpline.clear();
                        getline(eachfile,tmpline);
                        if(tmpline.size() > 0 && tmpline[tmpline.size()-1] == '\n')
                                tmpline[tmpline.size()-1] = ' ';
                        line = line + tmpline;                        
                        getline(eachfile,tmpline);
                }
                line = line + "\n";                
                data_vec.push_back(line);
                data_len.push_back(line.size());
                data_label.push_back(label);
                total += line.size(); 

                eachfile.close();
                eachfile.clear();   // this is important! this clear all the errorflags associated to eachfile
                
                i++;
                if(i%100 == 0) printf(".");                
        }
        printf(" [%d examples]",i);
        indatafile.close();
        datasize = data_vec.size();
        
        ofstream outfile;
        outfile.open(kernelmatrixfn.c_str());
        if(!outfile) 
        {
                printf("\nUnable to open output file <%s>\n", kernelmatrixfn.c_str());                
                exit(EXIT_FAILURE);
        }
        outfile.setf(ios::fixed);

                // self kernel evaluation of each string
        printf("\nComputing k(x,x) for all x");
        vector<Real> k_ii(datasize,0.0);
        for(UInt32 i=0; i<datasize; i++) 
        {                
                StringKernel sk((UInt32)data_vec[i].size(), (SYMBOL*)data_vec[i].c_str(), swf, swfParam);
                sk.Set_Lvs();
                sk.PrecomputeVal();
                Real kval = 0.0;
                sk.Compute_K((SYMBOL*)data_vec[i].c_str(), data_vec[i].size(), kval);
                k_ii[i] = kval;
                if(i%100 == 0) printf(".");
        }
        
        // allocated mem for kernel matrix
        km = new double[datasize*datasize];

        // Step 1: Loop for computing kernel matrix
        printf("\nComputing km rows");
        for(UInt32 i=0; i<datasize; i++) 
        {                
                StringKernel sk((UInt32)data_vec[i].size(), (SYMBOL*)data_vec[i].c_str(), swf, swfParam);
                sk.Set_Lvs();
                sk.PrecomputeVal();
    
                for(UInt32 j=i; j<datasize; j++) 
                {
                        if(i==j) 
                        {
                                km[i*datasize+j] = 1.0;      // k_ii[i]/sqrt(k_ii[i])/sqrt(k_ii[i]); 
                                continue;
                        }
                        Real kval = 0.0;
                        sk.Compute_K((SYMBOL*)data_vec[j].c_str(), data_vec[j].size(), kval);
                        km[i*datasize + j] = kval/sqrt(k_ii[i])/sqrt(k_ii[j]);
                        km[j*datasize + i] = km[i*datasize + j];
                }
                if(i%100 == 0) printf(".");
        }

        // put km into file
        // libsvm precomputed km file format:
        // <label> 0:ID 1:k_11 2:k_12 ...
        printf("\nWriting km into file");
        for(UInt32 i=0; i<datasize; i++) 
        {
                outfile << (int)data_label[i] << " 0:" << i+1;
                for(UInt32 j=0; j<datasize; j++)
                {
                        outfile << " " << j+1 << ":" << km[i*datasize+j];
                }
                outfile << endl;
                if(i%100 == 0) printf(".");
        }
        printf("\n");
        
        return EXIT_SUCCESS;
}
