#include <fstream>
#include "DataType.h"
#include "ErrorCode.h"
#include "ESA.h"
#include "StringKernel.h"
#include <sys/time.h>
#include <math.h>

using namespace std;

void help()
{
        printf("============================================================\n"
               "=    Compute string kernel matrix as plain kernel matrix   =\n"
               "============================================================\n"
               "Usage: ./line-plainkm  swf  param  dataset_fn  kernelmatrix_fn\n"
               "Options:\n"
               "swf (substring weighting functions)\n"
               "  0 - Constant Weight (a dummy parameter)\n"
               "  1 - Exponential Decay (lambda=0.5)\n"
               "  2 - K-Spectrum (k=5)\n"
               "  3 - Bounded Range (n=1)\n"
               "param (swf parameter)\n\n"
               "Note: \n"
               "1. Each line of dataset is consider as one string\n"               
               "2. A SENTINEL character ('\\n') is appended to each string\n\n"          
              );
}


int main(int argc, char *argv[])
{
        if(argc != 5) 
        { 
                help(); 
                exit(EXIT_FAILURE);
        }

        int swf = atoi(argv[1]);
        Real swfParam = atof(argv[2]);
        string datasetfn = string(argv[3]);        
        string kernelmatrixfn = string(argv[4]);
               
        UInt32 datasize = 0;

        vector<string> data_vec;
        vector<UInt32> data_len;        
        double *km = 0;

        ifstream indatafile;
        indatafile.open(datasetfn.c_str());

        if(!indatafile)
        {
                printf("Unable to open dataset <%s>\n", datasetfn.c_str());
                exit(EXIT_FAILURE);
        }

        printf("Reading data");    
        int total = 0;
        string line;
        UInt32 i = 0;
        while(!indatafile.eof())
        {                
                line.clear();
                getline(indatafile, line);
                if(line.size() < 1) continue;
                line = line + SENTINEL;
                data_vec.push_back(line);
                data_len.push_back(line.size());                
                total += line.size(); 
                i++;                
                if(i%100 == 0) printf(".");                
        }
        printf(" [%d examples]",i);
        indatafile.close();
        datasize = i;
        
        
        ofstream outfile;
        outfile.open(kernelmatrixfn.c_str());
        if(!outfile) 
        {
                printf("\nUnable to open output file <%s>\n", kernelmatrixfn.c_str());
                exit(EXIT_FAILURE);
        }
        outfile.setf(ios::fixed);

        // self kernel evaluation of each string
        printf("\nComputing k(x,x) for all x");
        vector<Real> k_ii(datasize,0.0);
        for(UInt32 i=0; i<datasize; i++) 
        {                
                StringKernel sk((UInt32)data_vec[i].size(), (SYMBOL*)data_vec[i].c_str(), swf, swfParam);
                sk.Set_Lvs();
                sk.PrecomputeVal();
                Real kval = 0.0;
                sk.Compute_K((SYMBOL*)data_vec[i].c_str(), data_vec[i].size(), kval);
                k_ii[i] = kval;
                if(i%100 == 0) printf(".");
        }
        
        // allocated mem for kernel matrix
        km = new double[datasize*datasize];

        // Step 1: Loop for computing kernel matrix
        printf("\nComputing km rows");
        for(UInt32 i=0; i<datasize; i++) 
        {                
                StringKernel sk((UInt32)data_vec[i].size(), (SYMBOL*)data_vec[i].c_str(), swf, swfParam);
                sk.Set_Lvs();
                sk.PrecomputeVal();
    
                for(UInt32 j=i; j<datasize; j++) 
                {
                        if(i==j) 
                        {
                                km[i*datasize+j] = 1.0;      // k_ii[i]/sqrt(k_ii[i])/sqrt(k_ii[i]); 
                                continue;
                        }
                        Real kval = 0.0;
                        sk.Compute_K((SYMBOL*)data_vec[j].c_str(), data_vec[j].size(), kval);
                        km[i*datasize + j] = kval/sqrt(k_ii[i])/sqrt(k_ii[j]);
                        km[j*datasize + i] = km[i*datasize + j];
                }
                if(i%100 == 0) printf(".");
        }

        // put km into file
        // plain km file format:
        // k_11 ... k_1n
        // ...
        // K_n1 ... k_nn
        printf("\nWriting km into file");
        for(UInt32 i=0; i<datasize; i++) 
        {
                for(UInt32 j=0; j<datasize; j++)
                {
                        outfile << km[i*datasize+j] << " ";
                }
                outfile << endl;
                if(i%100 == 0) printf(".");
        }
        printf("\n");
  
        return EXIT_SUCCESS;
}
